define([
  'backbone',
  './effects/SlideEffect'
],

function(Backbone, SlideEffect) {

  /**
   * Rendering the view and setting props required by StackNavigator.
   * @private
   * @ignore
   *
   * @param {View} view View to be rendered.
   * @param {StackNavigator} stackNavigator View StackNavigator instance.
   */
  function appendView(view, stackNavigator) {

    if (!view.__backStackRendered__) {

      // Setting ref to parent StackNavigator
      view.stackNavigator = stackNavigator;

      // Setting default destructionPolicy if it's not set
      if (typeof view.destructionPolicy === 'undefined') view.destructionPolicy = 'auto';

      // Setting default styles
      // view.$el.css({position:'absolute', visibility:'hidden', overflow:'hidden', width:'100%', height:'100%'});
      // view.$el.css({position:'absolute', visibility:'hidden'});
      stackNavigator.css = stackNavigator.css || {};
      stackNavigator.css.visibility = 'hidden';
      view.$el.css(stackNavigator.css);
      stackNavigator.$el.append(view.el);


    } else {
      // Resetting visibility to hidden
      view.$el.css({visibility:'hidden'});
      view.$el.css('display', 'none');
   
    }

    // Adding view to the DOM
    //stackNavigator.$el.append(view.el);
    //view.$el.css("z-index",1);
    //view.$el.css('opacity','1');
    //view.$el.css('visibility', 'visible');
    //       $toView.css('-webkit-transform', 'translate3d(0,0,-1px)');
    view.$el.css('display', 'block');
    
    if(view.destructionPolicy != 'never')
      stackNavigator.$el.append(view.el);

    if (!view.__backStackRendered__) {
      // Rendering the view
      view.render.call(view);

      // Setting default of __backStackRendered__ property
      view.__backStackRendered__ = true;
    }
  }

  /**
   * Creates event objects triggered by BackStack.
   * @private
   * @ignore
   *
   * @param {string} type Event type name.
   * @param {*} args Event args.
   * @param {boolean} cancelable Flag indicating if event is cancelable.
   * @return {event} The new object.
   */
  function createEvent(type, args, cancelable) {
    return _.extend({

      type:type,

      cancelable:_.isUndefined(cancelable) ? false : cancelable,

      preventDefault:function () {
        if (this.cancelable)
          this.isDefaultPrevented = function () {
            return true;
          };
      },

      isDefaultPrevented:function () {
        return false;
      },

      trigger:function (target) {
        target.trigger(this.type, this);
        return this;
      }
    }, args);
  }

  /**
   * Private common push method.
   * @private
   * @ignore
   *
   * @param {object} fromViewRef Reference to from view.
   * @param {object} toViewRef Reference to to view.
   * @param {number} replaceHowMany Number of views to replace with pushed view.
   * @param {Effect} transition Transition to played during push.
   */
  function push(fromViewRef, toViewRef, replaceHowMany, transition) {

    // Rendering view if required
    appendView(toViewRef.instance, this);

    transition = transition || this.defaultPushTransition || (this.defaultPushTransition = new SlideEffect({direction:'left'}));
    transition.play(fromViewRef ? fromViewRef.instance : null, toViewRef.instance,
      function () { // Callback function

        var remove = replaceHowMany > 0 ? this.viewsStack.splice(this.viewsStack.length - replaceHowMany, replaceHowMany)
          : (fromViewRef ? [fromViewRef] : null);

        if (remove != null) {
          _.each(remove, function (ref) {

            // Triggering viewDeactivate event
            createEvent('viewDeactivate', {target:ref.instance}).trigger(ref.instance);

            if (ref.instance.destructionPolicy == 'never') { // Detaching if destructionPolicy == 'never'
              //ref.instance.$el.detach();
            } else { // Removing if destructionPolicy == 'auto'
              ref.instance.$el.remove();
               // ref.instance = null;
            }
          }, this);
        }

        // Adding view to the stack internal array
        this.viewsStack.push(toViewRef);

        // Setting activeView property
        this.activeView = toViewRef.instance;

        // Triggering viewActivate event
        createEvent('viewActivate', {target:toViewRef.instance}).trigger(toViewRef.instance);

        // Triggering viewChanged event
        createEvent('viewChanged', {target:this}).trigger(this);

        // Popping item from actions queue
        popActionsQueue.call(this);

      }, this);
  }

  /**
   * Private common pop method.
   * @private
   * @ignore
   *
   * @param {object} fromViewRef Reference to from view.
   * @param {object} toViewRef Reference to to view.
   * @param {number} howMany Number of views to pop from the stack.
   * @param {Effect} transition Transition to played during pop.
   */
  function pop(fromViewRef, toViewRef, howMany, transition) {

    if (toViewRef) {
      // Recreating view instance if necessary
      toViewRef.instance = toViewRef.instance ? toViewRef.instance : new toViewRef.viewClass(toViewRef.options);
      // Rendering view if required
      appendView(toViewRef.instance, this);
    }

    transition = transition || this.defaultPopTransition || (this.defaultPopTransition = new SlideEffect({direction:'right'}));
    transition.play(fromViewRef.instance, toViewRef ? toViewRef.instance : null,
      function () { // Callback function

        // Popping views from a stack
        var remove = this.viewsStack.splice(this.viewsStack.length - howMany, howMany);
        _.each(remove, function (ref) {

          // Triggering viewDeactivate event
          createEvent('viewDeactivate', {target:ref.instance}).trigger(ref.instance);

          if (ref.instance.destructionPolicy == 'never') { // Detaching if destructionPolicy == 'never'
            //ref.instance.$el.detach();
             
          } else { // Removing if destructionPolicy == 'auto'
            ref.instance.$el.remove();
            //ref.instance = null;
          }
        }, this);

        if (toViewRef) { // If toViewRef exists activating it

          // Setting activeView property
          this.activeView = toViewRef.instance;

          // Triggering viewActivate event
          createEvent('viewActivate', {target:toViewRef.instance}).trigger(toViewRef.instance);

        } else { // Nulling activeView property
          this.activeView = null;
        }

        // Triggering viewChanged event
        createEvent('viewChanged', {target:this}).trigger(this);

        // Popping item from actions queue
        popActionsQueue.call(this);
      }, this);
  }
//I use transition more often so  swapped places with viewOptions
  function pushView(view, transition, viewOptions) {
    // Getting ref of the view on top of the stack
    var fromViewRef = _.last(this.viewsStack),
    // Creating new view instance if it is necessary
      toView = _.isFunction(view) ? new view(viewOptions) : view,
    // Creating new view ref
      toViewRef = {instance:toView, viewClass:toView.constructor, options:viewOptions},
    // Creating viewChanging event object and triggering it
      event = createEvent('viewChanging',
        {
          action:'push',
          fromViewClass:fromViewRef ? fromViewRef.viewClass : null,
          fromView:fromViewRef ? fromViewRef.instance : null,
          toViewClass:toViewRef.viewClass,
          toView:toViewRef.instance
        },
        true).trigger(this);

    // Checking if event wasn't cancelled
    if (event.isDefaultPrevented()) return null;

    push.call(this, fromViewRef, toViewRef, 0, transition);
  }

  function popView(transition) {
    if (this.viewsStack.length == 0) throw new Error('Popping from an empty stack!');

    // Getting ref of the view on top of the stack
    var fromViewRef = _.last(this.viewsStack),
    // Getting ref of the view below current one
      toViewRef = this.viewsStack.length > 1 ? this.viewsStack[this.viewsStack.length - 2] : null,
    // Creating viewChanging event object and triggering it
      event = createEvent('viewChanging',
        {
          action:'pop',
          fromViewClass:fromViewRef.viewClass,
          fromView:fromViewRef.instance,
          toViewClass:toViewRef ? toViewRef.viewClass : null,
          toView:toViewRef ? toViewRef.instance : null
        },
        true).trigger(this);

    // Checking if event wasn't cancelled
    if (event.isDefaultPrevented()) return;

    // Popping top view
    pop.call(this, fromViewRef, toViewRef, 1, transition);
  }

  function popAll(transition) {
    if (this.viewsStack.length == 0) throw new Error('Popping from an empty stack!');

    // Getting ref of the view on top of the stack
    var fromViewRef = _.last(this.viewsStack),
    // Creating viewChanging event object and triggering it
      event = createEvent('viewChanging',
        {
          action:'popAll',
          fromViewClass:fromViewRef.viewClass,
          fromView:fromViewRef.instance,
          toViewClass:null,
          toView:null
        },
        true).trigger(this);

    // Checking if event wasn't cancelled
    if (event.isDefaultPrevented()) return;

    // Popping top view
    pop.call(this, fromViewRef, null, this.viewsStack.length, transition);
  }

  function replaceView(view, viewOptions, transition) {
    if (this.viewsStack.length == 0) throw new Error('Replacing on an empty stack!');

    // Getting ref of the view on top of the stack
    var fromViewRef = _.last(this.viewsStack),
    // Creating new view instance if it is necessary
      toView = _.isFunction(view) ? new view(viewOptions) : view,
    // Creating new view ref
      toViewRef = {instance:toView, viewClass:toView.constructor, options:viewOptions},
    // Creating viewChanging event object and triggering it
      event = createEvent('viewChanging',
        {
          action:'replace',
          fromViewClass:fromViewRef.viewClass,
          fromView:fromViewRef.instance,
          toViewClass:toViewRef.viewClass,
          toView:toViewRef.instance
        },
        true).trigger(this);

    // Checking if event wasn't cancelled
    if (event.isDefaultPrevented()) return null;

    // Pushing new view on top
    push.call(this, fromViewRef, toViewRef, 1, transition);
  }

  function replaceAll(view, viewOptions, transition) {
    if (this.viewsStack.length == 0) throw new Error('Replacing on an empty stack!');

    // Getting ref of the view on top of the stack
    var fromViewRef = _.last(this.viewsStack),
    // Creating new view instance if it is necessary
      toView = _.isFunction(view) ? new view(viewOptions) : view,
    // Creating new view ref
      toViewRef = {instance:toView, viewClass:toView.constructor, options:viewOptions},
    // Creating viewChanging event object and triggering it
      event = createEvent('viewChanging',
        {
          action:'replaceAll',
          fromViewClass:fromViewRef.viewClass,
          fromView:fromViewRef.instance,
          toViewClass:toViewRef.viewClass,
          toView:toViewRef.instance
        },
        true).trigger(this);

    // Checking if event wasn't cancelled
    if (event.isDefaultPrevented()) return null;

    // Pushing new view on top
    push.call(this, fromViewRef, toViewRef, this.viewsStack.length, transition);
  }

  function popActionsQueue() {
    this.actionsQueue.splice(0, 1);
    if (this.actionsQueue.length > 0) {
      var action = this.actionsQueue[0],
        args = Array.prototype.slice.call(action.args);
      switch (action.fn) {
        case 'pushView':
          pushView.apply(this, args);
          break;
        case 'popView':
          popView.apply(this, args);
          break;
        case 'popAll':
          popAll.apply(this, args);
          break;
        case 'replaceView':
          replaceView.apply(this, args);
          break;
        case 'replaceAll':
          replaceAll.apply(this, args);
          break;
      }
    }
  }

  var StackNavigator = Backbone.View.extend(
    /** @lends BackStack.StackNavigator */
    {
      /**
       * @name StackNavigator#viewChanging
       * @event
       * @param {Object} e
       * @param {Boolean} [e.cancelable=true]
       */

      /**
       * An array with all the view refs on the stack.
       */
      viewsStack:null,

      /**
       * View on top of the stack.
       */
      activeView:null,

      /**
       * Default push transition effect.
       */
      defaultPushTransition:null,

      /**
       * Default pop transition effect.
       */
      defaultPopTransition:null,

      /**
       * Queue of actions to be executed on the stack.
       */
      actionsQueue:null,

      /**
       * Initializes StackNavigator.
       *
       * @param {Object} options This is a Backbone.View options hash that can have popTransition and pushTransition
       * properties that can be initiated for this instance of navigator.
       *
       * @constructs
       * */
      initialize:function (options) {
        // Setting default styles
        // this.$el.css({overflow:'hidden'});

        // Setting new viewsStack array
        this.viewsStack = [];

        // Setting new queue of actions
        this.actionsQueue = [];

        // Setting default pop transition
        if (options.popTransition) this.defaultPopTransition = options.popTransition;

        // Setting default push transition
        if (options.pushTransition) this.defaultPushTransition = options.pushTransition;

        // css to add to view when appending
        this.css = options.css || {}; 
      },

      /**
       * Pushes new view to the stack.
       *
       * @param {Backbone.View || Backbone.ViewClass} view View class or view instance to be pushed to the stack.
       * @param {Object} viewOptions Options to be passed if view is contructed by StackNavigator.
       * @param {Effect} transition Transition effect to be played when pushing new view.
       */
      pushView:function (view, transition, viewOptions) {
        // Pushing current action to the queue
        this.actionsQueue.push({fn:'pushView', args:arguments});

        if (this.actionsQueue.length == 1) pushView.call(this, view, transition, viewOptions);
      },

      /**
       * Pops an active view from a stack and displays one below.
       *
       * @param {Effect} transition Transition effect to be played when popping new view.
       */
      popView:function (transition) {
        
        // Pushing current action to the queue
        this.actionsQueue.push({fn:'popView', args:arguments});

        if (this.actionsQueue.length == 1) popView.call(this, transition);
      },

      /**
       * Pops all views from a stack and leaves empty stack.
       *
       * @param {Effect} transition Transition effect to be played when popping views.
       */
      popAll:function (transition) {
        // Pushing current action to the queue
        this.actionsQueue.push({fn:'popAll', args:arguments});

        if (this.actionsQueue.length == 1) popAll.call(this, transition);
      },

      /**
       * Replaces view on top of the stack, with the one passed as a view param.
       *
       * @param {Backbone.View || Backbone.ViewClass} view View class or view instance to be pushed on top of the stack instead of current one.
       * @param {Object} viewOptions Hash with options to be passed to the view, if view param is not an instance.
       * @param {Effect} transition Transition effect to be played when replacing views.
       */
      replaceView:function (view, viewOptions, transition) {
        // Pushing current action to the queue
        this.actionsQueue.push({fn:'replaceView', args:arguments});

        if (this.actionsQueue.length == 1) replaceView.call(this, view, viewOptions, transition);
      },

      /**
       * Replaces all of the views on the stack, with the one passed as a view param.
       *
       * @param {Backbone.View || Backbone.ViewClass} view View class or view instance to be pushed on top of the stack.
       * @param {Object} viewOptions Hash with options to be passed to the view, if view param is not an instance.
       * @param {Effect} transition Transition effect to be played when replacing views.
       */
      replaceAll:function (view, viewOptions, transition) {
        // Pushing current action to the queue
        this.actionsQueue.push({fn:'replaceAll', args:arguments});

        if (this.actionsQueue.length == 1) replaceAll.call(this, view, viewOptions, transition);
      }
    });

  return StackNavigator;
});